@@ -1,12 +1,82 @@
 #!/usr/bin/env python3
 """
 Module SETUPUTILS -- Utilities to automate boilerplate in Python setup scripts
-Copyright (C) 2012-2013 by Peter A. Donis
+Copyright (C) 2012-2022 by Peter A. Donis
 
 Released under the Python Software Foundation License.
 
-The purpose of this module is to help automate away much of the
-boilerplate that goes into Python setup scripts. Typical usage:
+This module was originally released to help automate away much of the
+boilerplate that goes into Python setup scripts. However, the tools
+for packaging Python applications and libraries have evolved a lot since
+then, and setuputils now supports a new mode of operation that takes
+advantage of the updated standards and tools.
+
+The preferred way to use setuputils now is to have it build a setup.cfg
+file from inputs that you supply and from its ability to auto-discover
+the things that typically go into a distribution of your Python application
+or library. The main tool that uses setup.cfg, setuptools, also has much
+the same capability, but the advantage with setuputils is that you get
+to see all the output of the autodiscovery logic in the final setup.cfg
+file *before* using it to build your distribution. With setuptools, you
+have no way of getting an advance look at what the tool thinks should go
+into your distribution; you only see what's in the distribution after it
+is built.
+
+Using setuputils in this mode is simple: for each section of setup.cfg,
+you put options that you want to specify in advance rather than have
+setuputils auto-discover in an .in file in the root of your source tree.
+For example, you would put metadata in the file metadata.in, options in
+the file options.in, etc. You would not include anything that you want
+setuputils to auto-discover, so, for example, if you want setuputils to
+auto-discover your packages, you would not include "packages" in your
+options.in file at all (whereas with setuptools you would include
+"packages = find:" in your setup.cfg).
+
+Once you have created your .in files, then you simply execute
+
+    python3 -m setuputils
+
+in the root of your source tree. This will build the setup.cfg file for
+you. You can then look at it to make sure it is correct before using a
+build backend to build your distribution.
+
+With this mode, if you are using a PEP 517 compliant build backend, you
+do not need a setup.py script at all. (Setuptools qualifies as such a
+backend as long as you include a pyproject.toml file and specify
+setuptools in it, as described in the Python packaging documentation.)
+If you do have a setup.py script, all it would need to contain is
+an import of ``setuptools`` and call to ``setuptools.setup()`` with no
+arguments (since all of the information needed is in setup.cfg). Note,
+however, that if you have a setup.py script that was used with previous
+versions of setuputils, you do not have to transition it to the new
+format using .in files all at once. Setuputils will read any global
+variables that are defined in your setup.py, as you would have done in
+previous setuputils versions, and include them in what it outputs to
+setup.cfg after processing them just as it would have in previous
+versions, so you can transition things incrementally if that works
+better for your project. (However, you should remove any calls to
+setup_vars in your setup.py script, leaving only the call to
+``setuptools.setup()`` with no arguments, since all the information
+it needs will be in setup.cfg.)
+
+(Note that extension modules are currently not supported in setup.cfg,
+so you cannot use setuputils in setup.cfg mode if your distribution
+contains any extension modules. You can still use it in "legacy" mode,
+described below.)
+
+This release of setuputils still supports the "legacy" mode of operation,
+where you include the variables setuputils needs in your setup.py file.
+However, this mode of operation is deprecated and may be removed in a
+future release (although it is likely that the Python packaging
+ecosystem will have to continue to support such "legacy" builds for
+quite a while yet). The disadvantage of this mode is that you have to
+include the setuputils.py file alongside your setup.py file in your
+source distribution, whereas in the new mode above you can just install
+setuputils as a normal Python library on your development machine,
+and only include the setup.cfg file it generates in your distribution.
+
+Typical usage in the original mode of operation was to put all of the
+information needed by setuputils in your setup.py file, thus:
 
     # declare variables here, for example...
     
@@ -19,15 +89,15 @@ boilerplate that goes into Python setup scripts. Typical usage:
     # declarations, or at least make them easier
     
     if __name__ == '__main__':
-        from distutils.core import setup
-        from setuputils import setup_vars
-        setup(**setup_vars(globals()))
+        from setuputils import setup_py_vars
+        from setuptools import setup
+        setup(**setup_py_vars(globals()))
 
-Or, you can take more fine-grained control over things by only
+Or, you could take more fine-grained control over things by only
 invoking particular sub-functions, for example:
 
     if __name__ == '__main__':
-        from distutils.core import setup
+        from setuptools import setup
         from setuputils import autodiscover_packages
         setup(
             name="myprog",
@@ -36,10 +106,10 @@ invoking particular sub-functions, for example:
             # other args
         )
 
-You can even mix the two methods, disabling general autodiscovery
+You could even mix the two methods, disabling general autodiscovery
 but using it for particular things:
     
-    from setuputils import autodiscover_packages, setup_vars
+    from setuputils import autodiscover_packages, setup_py_vars
     
     name = "myprog"
     
@@ -48,216 +118,59 @@ but using it for particular things:
     packages = autodiscover_packages(globals())
     
     if __name__ == '__main__':
-        from distutils.core import setup
-        setup(**setup_vars(globals(), autodiscover=False))
+        from setuptools import setup
+        setup(**setup_py_vars(globals(), autodiscover=False))
+
+Note that you call the ``setup_py_vars`` function in this mode
+(in older versions of setuputils it was just ``setup_vars``, but
+that is now an internal function used in both modes).
 
 See the docstrings of the individual functions below for more
-information.
+information on the various autodiscovery capabilities of
+setuputils.
 """
 
 import os
-from datetime import datetime
-from distutils.core import Extension
-from functools import partial
-from itertools import chain, dropwhile, islice, takewhile
 
 
-def _rst_header_lines(template, basename, **kwds):
-    # Generate header lines from template
-    if template:
-        return [
-            "{}{}".format(line, os.linesep) for line in
-            template.format(basename=basename, **kwds).splitlines()
-        ]
-    return []
+description_content_types = {
+    ".md": "text/markdown",
+    ".rst": "text/x-rst",
+    ".txt": "text/plain",
+}
 
 
-def _rst(basename, template, startline=None, rstext="rst", **kwds):
-    # Add header to file basename.rst and save to file basename
-    with open("{}.{}".format(basename, rstext), 'rU') as f:
-        lines = f.readlines()
-    outlines = _rst_header_lines(template, basename, **kwds) + lines[startline:]
-    with open(basename, 'w') as f:
-        f.writelines(outlines)
-
-
-def convert_rst(template=None,
-                dirname=".", rstnames=None, rstext="rst",
-                **kwds):
-    if rstnames is None:
-        rstnames = [
-            os.path.splitext(basename)[0]
-            for basename in os.listdir(dirname)
-            if os.path.splitext(basename)[1] == ".{}".format(rstext)
-        ]
-    for basename in rstnames:
-        _rst(basename, template=template, rstext=rstext, **kwds)
-
-
-def underline(line, underline='~'):
-    """Return an underline string for ``line``.
-    
-    Intended for use with level 3 and lower headings.
-    
-    Note that the default for ``underline`` assumes that the level
-    1 and 2 underlines are the standard ``=`` and ``-``. You can
-    pass a different underline character to this function if needed.
+def add_long_description(varmap, basename="README"):
+    """Add file reference for long description if available.
     """
     
-    if line.endswith('\n'):
-        line = line[:-1]
-    return "{}\n".format(underline * len(line))
+    for filename in os.listdir("."):
+        if filename.startswith(basename):
+            varmap['long_description'] = "file: {}".format(filename)
+            ext = os.path.splitext(filename)[-1]
+            content_type = description_content_types.get(ext)
+            if content_type:
+                varmap['long_description_content_type'] = content_type
+            return
 
 
-h3_prefix = '### '
-
-
-def _rst_from_md(basename, template=None, startline=None, mdext="md", **kwds):
-    # Convert Markdown text from file named basename.md to RST
-    # text saved to file named basename
-    with open("{}.{}".format(basename, mdext), 'rU') as f:
-        lines = f.readlines()
-    outlines = []
-    # Allow adding a header to the RST version (which can
-    # be filtered out when generating PKG_INFO if desired)
-    outlines.extend(_rst_header_lines(template, basename, **kwds))
-    # Normally you will want to pass an integer in as the
-    # startline keyword argument, since the Markdown file
-    # will probably have a title at the start, and the RST
-    # files used to provide PKG_INFO to PyPI should not have
-    # title lines (since PyPI adds a title itself)
-    for line in islice(lines, startline, None):
-        is_h3line = line.startswith(h3_prefix)
-        if is_h3line:
-            line = line[len(h3_prefix):]
-        outlines.append(line)
-        if is_h3line:
-            outlines.append(underline(line))
-    with open(basename, 'w') as f:
-        f.writelines(outlines)
-
-
-def convert_md_to_rst(template=None,
-                      dirname=".", mdnames=None, mdext="md",
-                      **kwds):
-    """Convert Markdown text files to RST text files.
-    
-    Useful for projects where "official" releases are on PyPI but
-    development source is visible somewhere else that uses Markdown
-    instead of ReStructuredText as its default text doc format,
-    such as github. The ``mdnames`` parameter gives the base names
-    of files to be converted; this defaults to every text file in
-    directory ``dirname`` (which defaults to the directory your
-    setup script is run from) with an ``.md`` extension.
+def add_pypi_url(varmap):
+    """Add the PyPI URL for program ``name`` from the setup vars.
     """
-    
-    if mdnames is None:
-        mdnames = [
-            os.path.splitext(basename)[0]
-            for basename in os.listdir(dirname)
-            if os.path.splitext(basename)[1] == ".{}".format(mdext)
-        ]
-    for basename in mdnames:
-        _rst_from_md(basename, template=template, mdext=mdext, **kwds)
-
-
-def current_date(fmt):
-    return datetime.now().strftime(fmt)
-
-
-def long_description(varmap, filename="README",
-                     startline=None, startspec=None,
-                     endline=None, endspec=None):
-    """Extract long description suitable for PyPI from RST text file.
-    
-    Useful to automate generating the ``long_description``
-    setup parameter. Looks in ``filename``, which should be
-    a ReStructuredText file, and returns all the text from
-    the given starting line (which is included in the result)
-    to the given ending line (which is not included in the
-    result). Lines can be specified either by line number
-    (``startline`` or ``endline``) or a string the line
-    starts with (``startspec`` or ``endspec``). Defaults
-    to using the ``README`` file.
-    """
-    
-    startline = startline or varmap.get('startline')
-    startspec = startspec or varmap.get('startspec')
-    endline = endline or varmap.get('endline')
-    endspec = endspec or varmap.get('endspec')
-    
-    with open(filename, 'rU') as f:
-        lines = f.readlines()
-    
-    if startline or endline:
-        innerlines = islice(lines, startline, endline)
-    else:
-        innerlines = lines
-    if startspec:
-        inner = dropwhile(
-            lambda line: not line.startswith(startspec),
-            innerlines
-        )
-    else:
-        inner = innerlines
-    if endspec:
-        desclines = takewhile(
-            lambda line: not line.startswith(endspec),
-            inner
-        )
-    else:
-        desclines = inner
-    
-    return "".join(desclines)
-
-
-def pypi_url(varmap):
-    """Return the PyPI URL for program ``name`` from the setup vars.
-    """
-    return "http://pypi.python.org/pypi/{}".format(varmap['name'])
-
-
-def provides(varmap):
-    """Return a minimal ``provides`` list from the setup vars.
-    """
-    
-    try:
-        return ["{} ({})".format(varmap['name'], varmap['version'])]
-    except KeyError:
-        return [varmap['name']]
+    name = varmap.get('name')
+    if name:
+        varmap['url'] = "http://pypi.org/project/{}".format(name)
 
 
 def add_vars(varmap):
-    """Automate adding ``long_description``, ``url``, and ``provides`` setup vars.
+    """Automate adding ``long_description`` and ``url`` setup vars.
     """
     
     if 'long_description' not in varmap:
-        varmap['long_description'] = long_description(varmap)
+        add_long_description(varmap)
     if 'url' not in varmap:
-        varmap['url'] = pypi_url(varmap)
-    if 'provides' not in varmap:
-        varmap['provides'] = provides(varmap)
-
-
-def convert_lists(varmap,
-                  listnames=('classifiers', 'requires', 'provides', 'obsoletes')):
-    """Convert long strings to lists of strings.
-    
-    Allows variable names in ``listnames`` to be specified as
-    long strings instead of lists, for easier typing. The
-    ``license_map`` and ``devstatus_trove`` strings below give
-    examples of the kind of long strings that can be used.
-    """
-    
-    listnames = varmap.get('listnames', listnames)
-    for listname in listnames:
-        try:
-            var = varmap[listname]
-        except KeyError:
-            pass
-        else:
-            if isinstance(var, str):
-                varmap[listname] = var.strip().splitlines()
+        add_pypi_url(varmap)
+    # provides format changed and it's not used anyway, so don't automatically include it
 
 
 def add_classifier_python(varmap):
@@ -328,20 +241,18 @@ def add_classifier_license(varmap):
     """Convert ``license`` setup var to Trove classifier.
     
     Allows short specification of common licenses while still
-    adhering to the PyPI spec, which says that the ``license``
-    setup keyword should only be used for licenses that don't
-    have a Trove classifier.
+    including the full Trove classifier as expected by PyPI.
     """
     
-    try:
-        c = license_map[varmap['license']]
-    except KeyError:
-        pass
-    else:
-        varmap.setdefault('classifiers', []).append(c)
-        # PyPI docs say the "License" keyword should only be used for
-        # licenses that don't have a Trove classifier
-        del varmap['license']
+    license = varmap.get('license')
+    if license:
+        c = license_map.get(license)
+        if c:
+            varmap.setdefault('classifiers', []).append(c)
+            # PyPI docs say the "License" keyword should only be used for
+            # licenses that don't have a Trove classifier, but nobody
+            # seems to pay attention, so we don't delete the 'license' key here
+            #del varmap['license']
 
 
 devstatus_trove = """
@@ -362,11 +273,8 @@ def add_classifier_dev_status(varmap):
     of having to type the full Trove classifier.
     """
     
-    try:
-        dev_status = varmap['dev_status']
-    except KeyError:
-        pass
-    else:
+    dev_status = varmap.get('dev_status')
+    if dev_status:
         for c in devstatus_trove:
             if c.endswith(dev_status):
                 varmap.setdefault('classifiers', []).append(c)
@@ -382,7 +290,9 @@ def add_classifiers(varmap):
     add_classifier_python(varmap)
     add_classifier_license(varmap)
     add_classifier_dev_status(varmap)
-    varmap['classifiers'] = sorted(varmap['classifiers'])
+    classifiers = varmap.get('classifiers')
+    if classifiers:
+        varmap['classifiers'] = sorted(classifiers)
 
 
 def package_root(varmap):
@@ -435,7 +345,13 @@ def add_py_modules(varmap):
     """
     
     if 'py_modules' not in varmap:
-        varmap['py_modules'] = autodiscover_modules(varmap)
+        modules = autodiscover_modules(varmap)
+        if modules:
+            varmap['py_modules'] = modules
+
+
+def marked(filenames):
+    return "__init__.py" in filenames
 
 
 def dir_to_package(dirname):
@@ -449,50 +365,96 @@ def package_paths(varmap):
     your actual setup script, but it can be useful for testing.
     It uses the "root" package, as determined from your setup
     vars, to determine where to start the search.
+    
+    Note that this function detects PEP 420 implicit namespace packages,
+    but only for a limited set of cases: it only includes directories that
+    do not have an __init__.py file but which have at least one subdirectory
+    that *does* have an __init__.py file. In other words, only implicit
+    namespace packages that have non-namespace packages inside them will
+    be detected. The PEP 420 spec allows implicit namespace packages "all the
+    way down", so to speak, with no __init__.py files anywhere, but trying
+    to auto-detect this, as setuptools does, either results in many false
+    positives (for example, scripts in a "scripts" directory of your source
+    distribution) or requires a "src" layout where all source code is put
+    inside a subdirectory of your distribution (which is fine if it works
+    for you, but might not be a good fit for all projects). For cases where
+    the setuptools logic works, you can use its autodetection instead of
+    the method here by putting "packages = find_namespace:" in your options.in
+    file instead of leaving out "packages" altogether.
     """
     
     rootdir = package_root(varmap)
-    result = []
+    packages = []
     for dirname, subdirs, filenames in os.walk(rootdir):
         if dirname != rootdir:
-            if "__init__.py" in filenames:
-                result.append(dirname.split(os.sep, 1)[1])
+            if marked(filenames) or any(marked(os.listdir(os.path.join(dirname, subdir))) for subdir in subdirs):
+                packages.append(dirname.split(os.sep, 1)[1])
             else:
                 # Don't recurse into subdirs in non-packages
                 subdirs[:] = []
-    return result
+    return packages
+
+
+def _translate_pkg_names(dirnames):
+    # Translate directory names to package names
+    return [
+        dir_to_package(dirname)
+        for dirname in dirnames
+    ]
 
 
 def autodiscover_packages(varmap):
     """Return list of Python packages in your distro.
     """
     
-    return [
-        dir_to_package(dirname)
-        for dirname in package_paths(varmap)
-    ]
+    return _translate_pkg_names(package_paths(varmap))
 
 
 def add_packages(varmap):
     """Automatically fill in the ``packages`` setup var.
+    
+    Note that "legacy" namespace packages (i.e., ones that do not use the
+    PEP 420 implicit namespace package feature and have an __init__.py
+    file) will need to be identified manually in the "namespace_packages"
+    variable in your options.in file; setuputils has no way of auto-discovering
+    these. (Note that this is true of the autodiscovery mechanism in setuptools
+    as well.) Any packages included in the "namespace_packages" variable will not
+    be included in the "packages" variable.
     """
     
     if 'packages' not in varmap:
-        varmap['packages'] = autodiscover_packages(varmap)
+        packages = autodiscover_packages(varmap)
+        if packages:
+            namespace_packages = varmap.get('namespace_packages', ())
+            packages = [p for p in packages if p not in namespace_packages]
+            if packages:
+                varmap['packages'] = packages
 
 
 def _package_data_paths(pathname, ext_srcdir):
     # Return list of package subdirectories that are not packages
-    # themselves (and are therefore presumed to contain package data)
+    # themselves (and are therefore presumed to contain package data).
+    # Implicit namespace packages are detected and filtered out of
+    # the list using the same logic as in ``package_paths`` above.
     return [
         "{}/*.*".format(subdir)
         for subdir in os.listdir(pathname)
-        if os.path.isdir(os.path.join(pathname, subdir))
+        if (subdir != "__pycache__")
         and (subdir != ext_srcdir)
-        and ("__init__.py" not in os.listdir(os.path.join(pathname, subdir)))
+        and os.path.isdir(os.path.join(pathname, subdir))
+        and not marked(os.listdir(os.path.join(pathname, subdir)))
+        and all(
+            not marked(os.listdir(os.path.join(pathname, subdir, subsub)))
+            for subsub in os.listdir(os.path.join(pathname, subdir))
+            if os.path.isdir(os.path.join(pathname, subdir, subsub))
+        )
     ]
 
 
+def package_to_dir(pkgname):
+    return pkgname.replace('.', os.sep)
+
+
 def autodiscover_package_data(varmap, ext_srcdir=""):
     """Return mapping of package data paths to file lists.
     
@@ -504,9 +466,9 @@ def autodiscover_package_data(varmap, ext_srcdir=""):
     
     ext_srcdir = varmap.get('ext_srcdir', ext_srcdir)
     return dict(
-        (dir_to_package(pathname), _package_data_paths(pathname, ext_srcdir))
-        for pathname in package_paths(varmap)
-        if _package_data_paths(pathname, ext_srcdir)
+        (pkgname, _package_data_paths(package_to_dir(pkgname), ext_srcdir))
+        for pkgname in varmap.get('packages', ())
+        if _package_data_paths(package_to_dir(pkgname), ext_srcdir)
     )
 
 
@@ -515,7 +477,9 @@ def add_package_data(varmap):
     """
     
     if 'package_data' not in varmap:
-        varmap['package_data'] = autodiscover_package_data(varmap)
+        package_data = autodiscover_package_data(varmap)
+        if package_data:
+            varmap['package_data'] = package_data
 
 
 def autodiscover_extensions(varmap,
@@ -539,15 +503,17 @@ def autodiscover_extensions(varmap,
     Note that this function assumes that only one extension
     "lives" in a given directory; multiple extensions in the
     same directory can't be autodiscovered using this mechanism.
+    
+    Note also that builds with extension modules are currently
+    not supported in setup.cfg, so setuputils will raise a
+    RuntimeError if it sees any extensions in that mode.
     """
     
-    rootdir = package_root(varmap)
-    result = []
-    try:
-        extnames = varmap['ext_names']
-    except KeyError:
-        pass
-    else:
+    extnames = varmap.get('ext_names')
+    if extnames:
+        from setuptools import Extension
+        rootdir = package_root(varmap)
+        result = []
         ext_srcdir = varmap.get('ext_srcdir', ext_srcdir)
         ext_exts = varmap.get('ext_exts', ext_exts)
         for extname in extnames:
@@ -565,7 +531,7 @@ def autodiscover_extensions(varmap,
             result.append(
                 Extension(extname, sources)
             )
-    return result
+        return result
 
 
 def add_extensions(varmap):
@@ -573,7 +539,9 @@ def add_extensions(varmap):
     """
     
     if 'ext_modules' not in varmap:
-        varmap['ext_modules'] = autodiscover_extensions(varmap)
+        ext_modules = autodiscover_extensions(varmap)
+        if ext_modules:
+            varmap['ext_modules'] = ext_modules
 
 
 def autodiscover_datafiles(varmap):
@@ -594,13 +562,11 @@ def autodiscover_datafiles(varmap):
     directory would be installed into ``share/myprog/examples``.
     """
     
-    result = []
-    try:
-        datadirs = varmap['data_dirs']
-    except KeyError:
-        pass
-    else:
-        pathprefix = "share/{}".format(varmap['name'])
+    datadirs = varmap.get('data_dirs')
+    name = varmap.get('name')
+    if datadirs and name:
+        result = {}
+        pathprefix = "share/{}".format(name)
         for datadir in datadirs:
             for dirname, subdirs, filenames in os.walk(datadir):
                 if filenames and ("." not in dirname):
@@ -612,10 +578,8 @@ def autodiscover_datafiles(varmap):
                     ]
                     if distfiles:
                         distdir = dirname.replace(os.sep, '/')
-                        result.append(
-                            ("{}/{}".format(pathprefix, distdir), distfiles)
-                        )
-    return result
+                        result["{}/{}".format(pathprefix, distdir)] = distfiles
+        return result
 
 
 def add_datafiles(varmap):
@@ -623,7 +587,19 @@ def add_datafiles(varmap):
     """
     
     if 'data_files' not in varmap:
-        varmap['data_files'] = autodiscover_datafiles(varmap)
+        data_files = autodiscover_datafiles(varmap)
+        if data_files:
+            varmap['data_files'] = data_files
+
+
+def add_entry_points(varmap):
+    """Automatically fill in the ``entry_points`` setup var.
+    """
+    
+    if 'entry_points' not in varmap:
+        entry_points = varmap.get('entry_points_file')
+        if entry_points:
+            varmap['entry_points'] = entry_points
 
 
 def autodiscover_scripts(varmap, dirname="scripts"):
@@ -649,7 +625,9 @@ def add_scripts(varmap):
     """
     
     if 'scripts' not in varmap:
-        varmap['scripts'] = autodiscover_scripts(varmap)
+        scripts = autodiscover_scripts(varmap)
+        if scripts:
+            varmap['scripts'] = scripts
 
 
 def autodiscover_all(varmap):
@@ -667,6 +645,7 @@ def autodiscover_all(varmap):
     add_package_data(varmap)
     add_extensions(varmap)
     add_datafiles(varmap)
+    add_entry_points(varmap)
     add_scripts(varmap)
 
 
@@ -774,7 +753,7 @@ def make_manifest_in(varmap):
     
     # This ensures that pyc files are left out in case the source
     # tree has them from testing
-    lines.append("recursive-exclude {} *.pyc\n".format(varmap['name']))
+    lines.append("recursive-exclude . *.pyc\n")
     
     # Read from the in.in file last so the user can override
     # any of the above (shouldn't need to but just in case)
@@ -790,50 +769,82 @@ def make_manifest_in(varmap):
         f.writelines(lines)
 
 
-# The type values in this dict are not currently used, but
-# may be used in future
-
-distutils_keywords = dict(
+setuptools_keywords = dict(
     name=str,
     version=str,
     description=str,
     long_description=str,
+    long_description_content_type=str,
     author=str,
     author_email=str,
     maintainer=str,
     maintainer_email=str,
     url=str,
     download_url=str,
-    packages=list,
-    py_modules=list,
-    scripts=list,
+    project_urls=(str, list),
+    include_package_data=bool,
+    packages=(str, list),
+    py_modules=(str, list),
+    namespace_packages=(str, list),
+    scripts=(str, list),
+    ext_package=str,
     ext_modules=list,
-    classifiers=list,
+    dev_status=str,
+    classifiers=(str, list),
     distclass=type,
     script_name=str,
     script_args=list,
     options=dict,
     license=str,
+    license_file=str,
+    license_files=(str, list),
     keywords=(str, list),
     platforms=(str, list),
     cmdclass=dict,
-    data_files=list,
-    package_dir=str,
+    data_dirs=list,
+    data_files=dict,
+    package_dir=(str, list),
     package_data=dict,
-    requires=list,
-    provides=list,
-    obsoletes=list
+    exclude_package_data=dict,
+    requires=(str, list),
+    provides=(str, list),
+    obsoletes=(str, list),
+    zip_safe=bool,
+    setup_requires=(str, list),
+    install_requires=(str, list),
+    extras_require=dict,
+    python_requires=str,
+    entry_points_file=str,
+    entry_points=dict,
+    eager_resources=(str, list),
+    dependency_links=(str, list),
+    tests_require=(str, list),
+    autodiscover=bool,
+    egg_base=str,
+)
+
+list_convert_keywords = tuple(
+    k for k, v in setuptools_keywords.items() if (v is list) or (v == (str, list))
 )
 
 
-def setup_vars(varmap, autodiscover=True, force_manifest_in=False):
-    """Return dict of keyword arguments for distutils setup function.
+def convert_lists(varmap,
+                  list_keys=list_convert_keywords):
+    """Convert long strings to lists of strings.
     
-    Normally the ``varmap`` argument will be the ``globals()``
-    of your setup script. The dictionary returned by this function
-    is filtered to only include valid setup arguments, so you can
-    safely pass your module globals without worrying about extra
-    variables declared there.
+    Allows variable names in ``list_keys`` to be specified as
+    long strings instead of lists, for easier typing.
+    """
+    
+    for key in list_keys:
+        var_type = setuptools_keywords[key]
+        var = varmap.get(key)
+        if var and isinstance(var, str) and ((var_type is list) or ('\n' in var)):
+            varmap[key] = var.strip().splitlines()
+
+
+def setup_vars(varmap, autodiscover=True, force_manifest_in=False):
+    """Return dict of setup variables from dict of input variables.
     
     The ``autodiscover`` argument determines whether arguments
     not explicitly declared in ``varmap`` will be auto-discovered
@@ -861,20 +872,234 @@ def setup_vars(varmap, autodiscover=True, force_manifest_in=False):
     """
     
     varmap = dict(varmap)  # so we don't mutate the original
-    
-    if 'name' not in varmap:
-        raise ValueError("You must supply a program name.")
+    convert_lists(varmap)
     
     if autodiscover:
         add_vars(varmap)
-        convert_lists(varmap)
         add_classifiers(varmap)
         autodiscover_all(varmap)
     
-    if force_manifest_in or os.path.isfile("MANIFEST.in.in"):
+    if force_manifest_in or ((force_manifest_in is not None) and os.path.isfile("MANIFEST.in.in")):
         make_manifest_in(varmap)
     
     return dict(
         (k, v) for k, v in varmap.items()
-        if k in distutils_keywords
+        if k in setuptools_keywords
     )
+
+
+def convert_file_values(varmap,
+                        file_marker="file:"):
+    
+    file_values = {k: v for k, v in varmap.items() if isinstance(v, str) and v.startswith(file_marker)}
+    if file_values:
+        for key, value in file_values.items():
+            filename = value[len(file_marker):].strip()
+            with open(filename, 'r') as f:
+                data = f.read()
+            varmap[key] = data
+
+
+def setup_py_vars(varmap, autodiscover=True, force_manifest_in=True):
+    """Wrapper to return vars for ``setup.py`` legacy mode.
+    
+    In this mode ``force_manifest_in`` defaults to ``True`` since
+    at a minimum we need to make sure ``setuputils.py`` is included
+    with the distribution. Also we need to convert any "file:"
+    values to the actual file data since that mode only works with
+    setup.cfg
+    """
+    
+    varmap = setup_vars(varmap, autodiscover=autodiscover, force_manifest_in=force_manifest_in)
+    convert_file_values(varmap)
+    return varmap
+
+
+setup_cfg_sections = (
+    ("metadata", (
+        "name",
+        "version",
+        "description",
+        "long_description",
+        "long_description_content_type",
+        "author",
+        "author_email",
+        "maintainer",
+        "maintainer_email",
+        "url",
+        "download_url",
+        "project_urls",
+        "license",
+        "license_files",
+        "keywords",
+        "classifiers",
+        "platforms",
+        "requires",
+        "provides",
+        "obsoletes",
+    )),
+    ("options", (
+        "zip_safe",
+        "setup_requires",
+        "install_requires",
+        "python_requires",
+        "eager_resources",
+        "dependency_links",
+        "tests_require",
+        "include_package_data",
+        "packages",
+        "package_dir",
+        "namespace_packages",
+        "py_modules",
+        "entry_points",
+        "scripts",
+    )),
+    ("options.extras_require", ()),
+    ("options.entry_points", ()),
+    ("options.package_data", ()),
+    ("options.exclude_package_data", ()),
+    ("options.data_files", ()),
+    ("egg_info", (
+        "egg_base",
+    )),
+)
+
+no_setup_cfg_keys = (
+    "ext_package",
+    "ext_modules",
+    "options",
+    "cmdclass",
+)
+
+
+def read_setup_py():
+    import setup
+    return {attr: getattr(setup, attr) for attr in dir(setup) if not attr.startswith('_')}
+
+
+def strtobool(value):
+    if value.lower() in ("false", "no", "0"):
+        return False
+    if value.lower() in ("true", "yes", "1"):
+        return True
+    raise ValueError("{} is not a valid boolean".format(repr(value)))
+
+
+def read_in_lines(lines,
+                  sep="=", listsep=",", comparison_ops=("==", ">", "<")):
+    
+    in_vars = {}
+    open_key = None
+    open_value = None
+    
+    def close_open_key():
+        nonlocal open_key, open_value
+        if open_key:
+            in_vars[open_key] = open_value
+            open_key = open_value = None
+    
+    for line in lines:
+        dangling = line.startswith(' ')
+        line = line.strip()
+        if not line:
+            continue
+        if (not dangling) and (sep in line) and all(s not in line for s in comparison_ops):
+            close_open_key()
+            key, value = (s.strip() for s in line.split(sep))
+            if value:
+                key_type = setuptools_keywords.get(key, type(value))
+                if key_type in (list, dict):
+                    raise RuntimeError("{} cannot be specified on a single line".format(key))
+                elif (key_type == (str, list)) and listsep not in value:
+                    raise RuntimeError("{} should be comma-separated list".format(key))
+                elif key_type is bool:
+                    value = strtobool(value)
+                in_vars[key] = value
+            else:
+                open_key = key
+                open_type = setuptools_keywords.get(key, list)
+                if open_type is dict:
+                    raise RuntimeError("Cannot specify extra section tables in .in file.")
+                elif (open_type is not list) and (open_type != (str, list)):
+                    raise RuntimeError("String values must be specified on single line in .in file.")
+                if open_type == (str, list):
+                    open_type = list
+                open_value = open_type()
+        else:
+            open_value.append(line)
+    close_open_key()
+    return in_vars
+
+
+def read_in_files():
+    varmap = {}
+    for section, _ in setup_cfg_sections:
+        filename = "{}.in".format(section)
+        if os.path.isfile(filename):
+            with open(filename, 'r') as f:
+                lines = f.readlines()
+            section_in = read_in_lines(lines)
+            if "." in section:
+                key = section.split(".", 1)[-1]
+                varmap[key] = section_in
+            else:
+                varmap.update(section_in)
+    return varmap
+
+
+def write_section_header(lines, section):
+    lines.append("[{}]\n".format(section))
+
+
+def write_item_lines(lines, item, value):
+    if isinstance(value, list):
+        lines.append("{} =\n".format(item))
+        for v in value:
+            lines.append("    {}\n".format(v))
+    else:
+        lines.append("{} = {}\n".format(item, value))
+
+
+def write_section_items(lines, item_vars):
+    for item, value in item_vars.items():
+        if value and not isinstance(value, dict):
+            write_item_lines(lines, item, value)
+
+
+def write_section_close(lines):
+    lines.append("\n")
+
+
+def write_out_file(cfg_vars):
+    lines = []
+    for section, items in setup_cfg_sections:
+        if items:
+            item_vars = {item: cfg_vars.get(item) for item in items}
+        else:
+            key = section.split(".", 1)[-1]
+            item_vars = cfg_vars.get(key)
+        if item_vars and isinstance(item_vars, dict):
+            write_section_header(lines, section)
+            write_section_items(lines, item_vars)
+            write_section_close(lines)
+    with open("setup.cfg", "w") as f:
+        f.writelines(lines)
+
+
+def write_setup_cfg(autodiscover=True):
+    if os.path.isfile('setup.py'):
+        varmap = read_setup_py()
+        #raise RuntimeError("Cannot have setup.py file in setup.cfg mode.")
+    else:
+        varmap = {}
+    varmap.update(read_in_files())
+    autodiscover = varmap.get('autodiscover', autodiscover)
+    cfg_vars = setup_vars(varmap, autodiscover=autodiscover, force_manifest_in=None)
+    forbidden_keys = [k for k in cfg_vars if k in no_setup_cfg_keys]
+    if forbidden_keys:
+        raise RuntimeError("Keys {} are not supported in setup.cfg.".format(", ".join(forbidden_keys)))
+    write_out_file(cfg_vars)
+
+
+if __name__ == '__main__':
+    write_setup_cfg()
